require 'mongoid'

module WhoisXMLAPI
  class Result
    include Mongoid::Document
    include Mongoid::Timestamps

    # https://blog.bigbinary.com/2016/02/15/rails-5-makes-belong-to-association-required-by-default.html
    belongs_to :whoisable, :polymorphic => true, optional: true

    field :domain
    field :record_created
    field :record_updated
    field :record_expires
    field :registrar
    field :nameservers, :type => Array
    field :data_error
    field :status

    has_one :registrant, :as => :contactable, :class_name => "WhoisXMLAPI::Contact", :autosave => true
    has_one :admin,      :as => :contactable, :class_name => "WhoisXMLAPI::Contact", :autosave => true
    has_one :tech,       :as => :contactable, :class_name => "WhoisXMLAPI::Contact", :autosave => true
    has_one :billing,    :as => :contactable, :class_name => "WhoisXMLAPI::Contact", :autosave => true

    # Actual indexing is performed in db_mongoid_extend.rake
    # index([
    #   [ :whoisable_id, Mongo::ASCENDING ],
    #   [ :whoisable_type, Mongo::ASCENDING ]
    # ])
    index({whoisable_id: 1, whoisable_type: 1})


    def parse(record)
      self.domain = record['domainName']
      self.data_error = record['dataError'] ? record['dataError'] : nil
      self.status = record['status'] ? record['status'] : nil

      if record['registryData']
        self.record_created = record['registryData']['createdDate']
        self.record_updated = record['registryData']['updatedDate']
        self.record_expires = record['registryData']['expiresDate']
      else
        res = WhoisXMLAPI::UnParsable.new
        res.parse(record)
        return res
      end

      self.registrar = record['registrarName']
      self.nameservers = record['nameServers'] ? record['nameServers']['hostNames'] : (record['registryData'] && record['registryData']['nameServers'] ? record['registryData']['nameServers']['hostNames'] : nil)
      registrant_data = record['registrant'] || (record['registryData'] ? record['registryData']['registrant'] : nil)

      if registrant_data.nil?
        res = WhoisXMLAPI::UnParsable.new
        res.parse(record)
        return res
      end

      # more "optional" data
      administrative_data = record['administrativeContact'] || (record['registryData'] ? record['registryData']['administrativeContact'] : nil)
      technical_data      = record['technicalContact'] || (record['registryData'] ? record['registryData']['technicalContact'] : nil)
      billing_data        = record['billingContact'] || (record['registryData'] ? record['registryData']['billingContact'] : nil)

      self.registrant = WhoisXMLAPI::Contact.selectively_create_from_hash(registrant_data.merge({'email' => record['contactEmail']})) if registrant_data
      self.admin      = WhoisXMLAPI::Contact.selectively_create_from_hash(administrative_data) if administrative_data
      self.tech       = WhoisXMLAPI::Contact.selectively_create_from_hash(technical_data) if technical_data
      self.billing    = WhoisXMLAPI::Contact.selectively_create_from_hash(billing_data) if billing_data
    end


    def complete?
      ( self.domain and self.registrar and ( self.registrant or self.admin or self.tech) )
    end


    def private?
      %w/networksolutionsprivateregistration.com domainprivacygroup.com/.each do |priv|
        return true if self.registrant.email =~ /#{priv}/
      end
      false
    end


    def contact_email(contact)
      case contact
      when :registrant
        registrant ? registrant.email : nil
      when :admin
        admin ? admin.email : nil
      when :tech
        tech ? tech.email : nil
      when :billing
        billing ? billing.email : nil
      else
        nil
      end
    end


    def contact_email_host(contact)
      email = contact_email(contact)
      host = email.squish.split("@").last.downcase if email.present? && email.include?("@")
      PublicSuffix.valid?(host) ? host : nil
    end

  end
end
